defmodule ClickHouse.Format.Values do
  @moduledoc """
  An implementation of the ClickHouse `"Values"` format.
  """

  @behaviour ClickHouse.Format

  alias ClickHouse.{FormatError, Query}

  ################################
  # ClickHouse.Format Callbacks
  ################################

  @impl ClickHouse.Format
  @spec decode(ClickHouse.Result.t()) :: {:error, ClickHouse.FormatError.t()}
  def decode(_result), do: error("Decode not implemented.")

  @impl ClickHouse.Format
  @spec encode(ClickHouse.Query.t()) ::
          {:ok, ClickHouse.Query.t()} | {:error, ClickHouse.FormatError.t()}
  def encode(%Query{command: :insert, params_count: count}) when count > 1 do
    error("Values insert statement can only support a single parameter.")
  end

  def encode(%Query{command: :insert, params_count: 0, params: [_ | _]}) do
    error("Values insert statement with params must include a '$1'.")
  end

  def encode(%Query{command: :insert, params_count: 1, params: params} = query) do
    data = do_encode(params)
    encoded = String.replace(query.statement, "$1", data)
    {:ok, Query.put_encoded(query, encoded)}
  end

  ################################
  # Private API
  ################################

  defp do_encode(params) do
    Enum.map_join(params, &encode_row/1)
  end

  defp encode_row(values) do
    row = Enum.map_join(values, ",", &encode_value/1)
    "(" <> row <> ")"
  end

  defp encode_value(nil), do: "NULL"
  defp encode_value(false), do: "0"
  defp encode_value(true), do: "1"
  defp encode_value(integer) when is_integer(integer), do: to_string(integer)

  defp encode_value(%Date{} = date) do
    date |> to_string() |> with_quotes()
  end

  defp encode_value(string) when is_binary(string) do
    string |> escape() |> with_quotes()
  end

  defp encode_value(atom) when is_atom(atom) do
    atom |> to_string() |> escape() |> with_quotes()
  end

  defp encode_value(%DateTime{} = datetime) do
    datetime
    |> DateTime.truncate(:second)
    |> DateTime.to_iso8601()
    |> String.replace("Z", "")
    |> with_quotes()
  end

  defp encode_value(list) when is_list(list) do
    values = Enum.map_join(list, ",", &encode_value/1)
    "[" <> values <> "]"
  end

  defp with_quotes(string) do
    "'" <> string <> "'"
  end

  defp escape(string) do
    string
    |> String.replace("'", "\\'")
    |> String.replace(~s("), ~s(\\"))
  end

  defp error(message) do
    error = %FormatError{message: message}
    {:error, error}
  end
end
